import hashlib
import oss2
from urllib.parse import urlparse
import os
import datetime
from scrapy.exceptions import DropItem

class OSSPipeline(object):

    def __init__(self, access_key_id, access_key_secret, endpoint, bucket_name):
        self.access_key_id = access_key_id
        self.access_key_secret = access_key_secret
        self.endpoint = endpoint
        self.bucket_name = bucket_name

    def get_hash(self, url):
        return hashlib.md5(url.encode('utf8')).hexdigest()

    def process_item(self, item, spider):
        unique_id = item['unique_id']
        if not unique_id:
            raise DropItem('unique_id is None')

        oss_filename = item.get('oss_filename')
        # to maintain rule data in item, execute deepcopy method before to_json_str
        result = self.bucket.put_object(oss_filename, item.deepcopy().to_json_str(), headers={
            'Content-Type': 'application/json',
        })
        spider.logger.info(
            f"upload data to oss: {oss_filename}, url: {item.get('url', default='')}, status: {result.status}")
        return item

    def open_spider(self, spider):
        auth = oss2.Auth(self.access_key_id, self.access_key_secret)
        self.bucket = oss2.Bucket(auth, self.endpoint, self.bucket_name)
        spider.logger.info('connnect and auth oss success')

    def close_spider(self, spider):
        spider.logger.info(f'spider {spider.name} oss pipeline closed')

    @classmethod
    def from_crawler(cls, crawler):
        if crawler.spider is None:
            return
        OSS_ACCESS_KEY_ID = crawler.spider.settings.get('OSS_ACCESS_KEY_ID') if crawler.spider.settings.get(
            'OSS_ACCESS_KEY_ID') else os.environ.get('OSS_ACCESS_KEY_ID')
        OSS_ACCESS_KEY_SECRET = crawler.spider.settings.get('OSS_ACCESS_KEY_SECRET') if crawler.spider.settings.get(
            'OSS_ACCESS_KEY_SECRET') else os.environ.get('OSS_ACCESS_KEY_SECRET')
        OSS_END_POINT = crawler.spider.settings.get('OSS_END_POINT') if crawler.spider.settings.get(
            'OSS_END_POINT') else os.environ.get('OSS_END_POINT')
        OSS_BUCKET_NAME = crawler.spider.settings.get('OSS_BUCKET_NAME') if crawler.spider.settings.get(
            'OSS_BUCKET_NAME') else os.environ.get('OSS_BUCKET_NAME')
        return cls(
            OSS_ACCESS_KEY_ID,
            OSS_ACCESS_KEY_SECRET,
            OSS_END_POINT,
            OSS_BUCKET_NAME
        )



# -*- coding: utf-8 -*-

from scrapy.exceptions import DropItem
import pika
import os
import json
from scrapy_nc.queue.producer import Producer
class AMQPPipeline(object):
    def __init__(self,
                 mq_dev_user,
                 mq_dev_password,
                 mq_dev_host,
                 mq_dev_port,
                 mq_dev_vhost,
                 mq_prod_user,
                 mq_prod_password,
                 mq_prod_host,
                 mq_prod_port,
                 mq_prod_vhost):
        self.producer_dev = Producer(
            mq_dev_host, mq_dev_port, mq_dev_vhost, mq_dev_user, mq_dev_password)
        self.producer_prod = Producer(
            mq_prod_host, mq_prod_port, mq_prod_vhost, mq_prod_user, mq_prod_password)

    def process_item(self, item, spider):
        cralab_env = os.environ.get('CRAWLAB_ENV')
        if cralab_env == 'HK':
            return item
        queue_names = item.queue_names()
        if len(queue_names) == 0:
            spider.logger.info(
                f"queue name length is 0, item url {item.get('url')}")
            # raise DropItem(f"empty queue_name item {item.get('url')}")
            return item
        for queue_name in queue_names:
            tasks = []
            if queue_name.endswith('.dev'):
                tasks.append({
                    "env": "dev",
                    "queue_name": queue_name
                })
            elif queue_name.endswith('.prod'):
                tasks.append({
                    "env": "prod",
                    "queue_name": queue_name
                })
            else:
                tasks.append({
                    "env": "dev",
                    "queue_name": queue_name + '.dev'
                })
                tasks.append({
                    "env": "prod",
                    "queue_name": queue_name + '.prod'
                })
            for task in tasks:
                env = task['env']
                queue_name = task['queue_name']

                amqp_queue_use_item = spider.settings.get(
                    'AMQP_QUEUE_USE_ITEM')
                queue_data = {}
                if amqp_queue_use_item == True:
                    data = item
                    queue_data = data.to_json_str()
                else:
                    queue_data = json.dumps({
                        "filename": item.get('oss_filename'),
                        "channel": spider.name,
                        "source_id": item.get('unique_id'),
                        "url": item.get('url', default=''),
                    }, ensure_ascii=False)

                if env == 'dev':
                    try:
                        self.producer_dev.publish(queue_name, queue_data)
                    except:
                        pass
                else:
                    self.producer_prod.publish(queue_name, queue_data)
                    spider.logger.info(f"send to amqp {queue_name} item: {item['unique_id']}, url: {item.get('url', default='')}")
        return item

    def open_spider(self, spider):
        self.producer_dev.connect()
        self.producer_prod.connect()
        spider.logger.info(f'create channel success')

    def close_spider(self, spider):
        self.producer_dev.close()
        self.producer_prod.close()
        spider.logger.info(f'close connection success')

    @classmethod
    def from_crawler(cls, crawler):
        if crawler.spider is None:
            return
        mq_dev_user = crawler.spider.settings.get('MQ_DEV_USER') if crawler.spider.settings.get(
            'MQ_DEV_USER') else os.environ.get('MQ_DEV_USER')
        mq_dev_password = crawler.spider.settings.get('MQ_DEV_PASSWORD') if crawler.spider.settings.get(
            'MQ_DEV_PASSWORD') else os.environ.get('MQ_DEV_PASSWORD')
        mq_dev_host = crawler.spider.settings.get('MQ_DEV_HOST') if crawler.spider.settings.get(
            'MQ_DEV_HOST') else os.environ.get('MQ_DEV_HOST')
        mq_dev_port = crawler.spider.settings.get('MQ_DEV_PORT') if crawler.spider.settings.get(
            'MQ_DEV_PORT') else os.environ.get('MQ_DEV_PORT')
        mq_dev_vhost = crawler.spider.settings.get('MQ_DEV_VHOST') if crawler.spider.settings.get(
            'MQ_DEV_VHOST') else os.environ.get('MQ_DEV_VHOST')

        mq_prod_user = crawler.spider.settings.get('MQ_PROD_USER') if crawler.spider.settings.get(
            'MQ_PROD_USER') else os.environ.get('MQ_PROD_USER')
        mq_prod_password = crawler.spider.settings.get('MQ_PROD_PASSWORD') if crawler.spider.settings.get(
            'MQ_PROD_PASSWORD') else os.environ.get('MQ_PROD_PASSWORD')
        mq_prod_host = crawler.spider.settings.get('MQ_PROD_HOST') if crawler.spider.settings.get(
            'MQ_PROD_HOST') else os.environ.get('MQ_PROD_HOST')
        mq_prod_port = crawler.spider.settings.get('MQ_PROD_PORT') if crawler.spider.settings.get(
            'MQ_PROD_PORT') else os.environ.get('MQ_PROD_PORT')
        mq_prod_vhost = crawler.spider.settings.get('MQ_PROD_VHOST') if crawler.spider.settings.get(
            'MQ_PROD_VHOST') else os.environ.get('MQ_PROD_VHOST')

        return cls(
            mq_dev_user,
            mq_dev_password,
            mq_dev_host,
            mq_dev_port,
            mq_dev_vhost,
            mq_prod_user,
            mq_prod_password,
            mq_prod_host,
            mq_prod_port,
            mq_prod_vhost
        )



import os
from pymongo import MongoClient

MONGO_HOST = os.environ.get('CRAWLAB_MONGO_HOST')
MONGO_PORT = int(os.environ.get('CRAWLAB_MONGO_PORT', '27017'))
MONGO_DB = os.environ.get('CRAWLAB_MONGO_DB')
MONGO_USERNAME = os.environ.get('CRAWLAB_MONGO_USERNAME')
MONGO_PASSWORD = os.environ.get('CRAWLAB_MONGO_PASSWORD')
MONGO_AUTHSOURCE = os.environ.get('CRAWLAB_MONGO_AUTHSOURCE')

mongo_conn = MongoClient(
    host=MONGO_HOST,
    port=MONGO_PORT,
    username=MONGO_USERNAME,
    password=MONGO_PASSWORD,
    authSource=MONGO_AUTHSOURCE,
) if MONGO_HOST else None

mongo_db = mongo_conn.get_database(MONGO_DB) if mongo_conn else None



from urllib import parse
import hashlib

# 合并相对URL，并且处理掉 ../ 类似的地址
def process_url(referer, url):
    url = parse.urljoin(referer, url)
    u = parse.urlparse(url)
    _u = parse.urlparse(parse.urljoin(f'{u.scheme}://{u.netloc}', u.path))
    return parse.ParseResult(u.scheme, u.netloc, _u.path, u.params, u.query, u.fragment).geturl()


def get_hash(data_str):
    return hashlib.md5(data_str.encode('utf8')).hexdigest()



import logging
import json
import pika

class Producer:

    def __init__(self, host, port, virtual_host, username, password):
        if virtual_host == '':
            virtual_host = '/'
        self._is_enable = host and len(host) > 0
        if self._is_enable:
            self._params = pika.connection.ConnectionParameters(
                host=host,
                port=port,
                virtual_host=virtual_host,
                heartbeat=0,
                credentials=pika.credentials.PlainCredentials(username, password))
            self.queue_names = []

        self._conn = None
        self._channel = None

    def connect(self):
        if self._is_enable and ( not self._conn or self._conn.is_closed) :
            self._conn = pika.BlockingConnection(self._params)
            self._channel = self._conn.channel()

    def _publish(self, queue_name, msg):

        if not queue_name in self.queue_names:
            self._channel.queue_declare(queue_name,
                                            passive=False,
                                            durable=True,
                                            exclusive=False,
                                            auto_delete=False,
                                            arguments=None,
                                            )
            self.queue_names.append(queue_name)
        self._channel.basic_publish('',
                                    queue_name,
                                    msg,
                                    pika.BasicProperties(
                                            content_type='application/json',
                                            delivery_mode=2),
                                    )


    def publish(self, queue_name, msg):
        """Publish msg, reconnecting if necessary."""
        if self._is_enable:
            try:
                self._publish(queue_name,msg)
                logging.info(f'send to queue {queue_name}')
            except pika.exceptions.ConnectionClosed:
                logging.debug('reconnecting to queue')
                self.connect()
                self._publish(queue_name,msg)

    def close(self):
        if self._conn and self._conn.is_open:
            logging.debug('closing queue connection')
            self._conn.close()




from scrapy.logformatter import LogFormatter
import logging
import os

class DefaultLogFormatter(LogFormatter):
    def dropped(self, item, exception, response, spider):
        return {
            'level': logging.INFO,
            'msg': u"Dropped: %(exception)s" + os.linesep,
            'args': {
                'exception': exception,
                'item': item,
            },
        }


default_log_formatter = 'scrapy_nc.log_formatter.DefaultLogFormatter'



from .amqp_pipeline import AMQPPipeline
from .oss_pipeline import OSSPipeline
from .mongo_pipeline import MongoPipeline
from .item_check_pipeline import ItemCheckPipeline
from .filter_pipeline import FilterCheckPipeline, FilterSavePipeline
# from .redis_duplicate_pipeline import RedisDuplicatesPipeline

ITEM_CHECK_PIPELINE = 'scrapy_nc.pipelines.ItemCheckPipeline'
FILTER_CHECK_PIPELINE = 'scrapy_nc.pipelines.FilterCheckPipeline'
OSS_PIPELINE = 'scrapy_nc.pipelines.OSSPipeline'
AMQP_PIPELINE = 'scrapy_nc.pipelines.AMQPPipeline'
MONGO_PIPELINE = 'scrapy_nc.pipelines.MongoPipeline'
FILTER_SAVE_PIPELINE = 'scrapy_nc.pipelines.FilterSavePipeline'

DEFAULT_PIPELINES = {
    ITEM_CHECK_PIPELINE: 500,
    FILTER_CHECK_PIPELINE: 550,
    OSS_PIPELINE: 600,
    AMQP_PIPELINE: 650,
    MONGO_PIPELINE: 700,
    FILTER_SAVE_PIPELINE: 750,
}



import scrapy
import json
from scrapy_nc.item import BaseItem

class NewsItem(BaseItem):
    url = scrapy.Field()
    title = scrapy.Field()
    author = scrapy.Field()
    content = scrapy.Field()
    html = scrapy.Field()
    picture = scrapy.Field()
    summary = scrapy.Field()
    publish_date = scrapy.Field()
    crawled_date = scrapy.Field()
    language = scrapy.Field()
    channel = scrapy.Field()
    source_type = scrapy.Field()
    source = scrapy.Field()
    extra = scrapy.Field()
    rule = scrapy.Field()
    message_package = scrapy.Field()

    def to_dict(self):
        dt = self.__dict__['_values']
        # del rule data before serialize
        dt.pop('rule', '')
        dt.pop('message_package', '')
        return dt

    def to_json_str(self):
        if self.get('publish_date') is not None:
            self['publish_date'] = self['publish_date'].isoformat()
        if self['crawled_date'] is not None:
            self['crawled_date'] = self['crawled_date'].isoformat()
        return json.dumps(self.to_dict(), ensure_ascii=False)


from redisbloom.client import Client
from redis.exceptions import ResponseError
import logging
import os

logger = logging.getLogger('filter')
logger.setLevel(logging.INFO)

HOST = os.environ.get('BLOOM_FILTER_HOST', None)
PORT = int(os.environ.get('BLOOM_FILTER_PORT', '36379'))
filter_name = 'bloom'

filter = None
if HOST:
    filter = Client(host=HOST, port=PORT)
    logger.info('connect filter success')
else:
    logger.info('not found bloom filter env, skip init filter')

if filter:
    try:
        filter.bfCreate(filter_name, 0.0001, 1000000000)
    except ResponseError as e:
        if e.args[0] != 'item exists':
            raise e
        logger.info(f'filter exists {filter_name}')
else:
    logger.info('not found filter, sikp init filter')


def build_filter_data(spider_name, unique_id):
    return spider_name + unique_id


# check data for existence
def check(spider_name, unique_id):
    return filter.bfExists(
        filter_name, build_filter_data(spider_name, unique_id))


# save data to filter
def save(spider_name, unique_id):
    filter.bfAdd(filter_name, build_filter_data(spider_name, unique_id))



from scrapy_nc.filter import filter, build_filter_data
from scrapy_nc.filter import filter_name
from scrapy.exceptions import DropItem


class FilterSavePipeline(object):

    def process_item(self, item, spider):
        unique_id = item.get('unique_id')
        filter.bfAdd(filter_name, build_filter_data(spider.name, unique_id))
        spider.logger.info(f'save to filter {unique_id} {spider.name}')
        return item


class FilterCheckPipeline(object):
    def process_item(self, item, spider):
        unique_id = item.get('unique_id')
        if filter.bfExists(filter_name, build_filter_data(spider.name, unique_id)):
            raise DropItem(f"exists item {spider.name} {item['unique_id']}")
        return item



from scrapy.exceptions import DropItem
import datetime


class ItemCheckPipeline(object):

    def process_item(self, item, spider):
        unique_id = item.get('unique_id')
        if not unique_id:
            spider.logger.error(f'unique_id is None')
            raise DropItem('unique_id is None')

        queues = item.queue_names()
        if len(queues) == 0:
            spider.logger.warn(f'queue_names length is 0')
        item['crawled_at'] = datetime.datetime.utcnow().replace(
                        tzinfo=datetime.timezone.utc).astimezone().isoformat()

        oss_filename = item.get('oss_filename')
        if not oss_filename:
            if unique_id.find('/') != -1:
                msg = 'unique_id format error, find / in unicode_id'
                spider.logger.error(msg)
                raise DropItem(msg)
            oss_filename = f'{spider.name}/{unique_id}'
            item['oss_filename'] = oss_filename
        return item



from scrapy_nc.crawlab import get_task_id
from scrapy_nc.db import mongo_db
import json


class MongoPipeline(object):
    def process_item(self, item, spider):
        if not self.spider_collection:
            spider.logger.warn(f'mongodb not collected, ignore save')
            return item
        ensure_unique_index = spider.settings.get('ENSURE_UNIQUE_INDEX')
        my_item = item.deepcopy()
        if my_item.get('html'):
            my_item['html'] = None
        item_dict = dict(my_item)
        item_dict['task_id'] = get_task_id()
        if ensure_unique_index:
            self.spider_collection.update_one({
                'unique_id':   my_item.get('unique_id')
            }, {"$set": item_dict}, upsert=True)
        else:
            self.spider_collection.update_one({
                'unique_id':   my_item.get('unique_id'),
                'task_id':   my_item.get('task_id')
            }, {"$set": item_dict}, upsert=True)
        return item

    def open_spider(self, spider):
        spider.logger.info('open_spider')
        if not mongo_db:
            self.spider_collection = None
            spider.logger.error('mongodb not configed')
            return

        self.spider_collection = mongo_db.get_collection(spider.name)
        ensure_unique_index = spider.settings.get('ENSURE_UNIQUE_INDEX')

        res = json.dumps(self.spider_collection.index_information())
        spider.logger.info(
            f'index_information {res}')
        if ensure_unique_index:
            index_name = 'unique_id'
            if index_name not in self.spider_collection.index_information():
                self.spider_collection.create_index(
                    'unique_id', unique=True, name=index_name)
        else:
            index_name = 'unique_id_task_id'
            if index_name not in self.spider_collection.index_information():
                self.spider_collection.create_index(
                    [('unique_id', 1), ('task_id', 1)], unique=True, name=index_name)


from setuptools import setup, find_packages
import io
import scrapy_nc

def read_file(filename):
    with io.open(filename) as fp:
        return fp.read().strip()

def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]

setup(
    name='scrapy_nc',
    version='0.1.2',
    url='https://github.com/fantasy/scrapy_nc',
    description='Scrapy plugins in NoCode',
    author='fantasy614@nocode.com',
    packages=['scrapy_nc.pipelines',
              'scrapy_nc.item',
              'scrapy_nc.db',
              'scrapy_nc.crawlab',
              'scrapy_nc.middlewares',
              'scrapy_nc.filter',
              'scrapy_nc.helper',
              'scrapy_nc.queue',
              'scrapy_nc.log_formatter'
              ],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
        'Framework :: Scrapy',
    ],
    install_requires=['scrapy'],
    requires=['scrapy (>=0.24.5)'],
)



import scrapy
import json


class BaseItem(scrapy.Item):
    crawled_at = scrapy.Field()
    oss_filename = scrapy.Field()
    unique_id = scrapy.Field()

    def to_dict(self):
        dt = self.__dict__['_values']
        return dt

    def to_json_str(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def queue_names(self):
        return []


# from scrapy.exceptions import DropItem
# from pyreBloom import pyreBloom

# class RedisDuplicatesPipeline(object):
#     def __init__(self, redis_host, redis_port, redis_password, redis_db):
#         self.redis_host = redis_host
#         self.redis_port = redis_port
#         self.redis_password = redis_password
#         self.redis_db = redis_db

#     def process_item(self, item, spider):
#         unique_key = item.unique_key()
#         if self.bloom.contains(unique_key):
#             raise DropItem("Duplicate item found: %s" % item)
#         else:
#             self.bloom.add(unique_key)
#             return item

#     def open_spider(self, spider):

#         redis_conf = {
#             'host': self.redis_host,
#             'password': self.redis_password,
#             'port': self.redis_port,
#             'db': self.redis_db
#         }

#         self.bloom = pyreBloom('bloom' + spider.name,
#                                spider.settings.MAX_COUNT, 0.0001, **redis_conf)

#     def close_spider(self, spider):
#         self.bloom.close()
#         spider.logger.info(f'close connection success')

#     @classmethod
#     def from_crawler(cls, crawler):
#         return cls(
#             crawler.spider.settings['REDIS_HOST'],
#             crawler.spider.settings['REDIS_PORT'],
#             crawler.spider.settings['REDIS_PASSWORD'],
#             crawler.spider.settings['REDIS_DB'],
#         )



import os
import requests
import json


class ProxyDownloaderMiddleware(object):

    def __init__(self, proxy_server_addr):
        self.proxy_server_addr = proxy_server_addr

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        proxy_server_addr = crawler.spider.settings.get('PROXY_SERVER_ADDR') if crawler.spider.settings.get(
            'PROXY_SERVER_ADDR') else os.environ.get('PROXY_SERVER_ADDR')
        s = cls(proxy_server_addr)
        return s

    def process_request(self, request, spider):
        if request.meta['skip_proxy'] == True:
            return None
        req_url = request.url
        if 'splash' in request.meta:
            req_url = request.meta['splash']['args']['url']
        proxy = self.get_proxy(request, spider.name, req_url)
        if proxy is not None:
            if proxy['pass'] is not None and proxy['pass'] != '':
                proxy_server = f"http://{proxy['user']}:{proxy['pass']}@{proxy['host']}:{proxy['port']}"
            else:
                proxy_server = f"http://{proxy['host']}:{proxy['port']}"

            if 'splash' in request.meta:
                request.meta['splash']['args']['proxy'] = proxy_server
            else:
                request.meta['proxy'] = proxy_server
                request.headers["Proxy-Authorization"] = proxy['basic_auth_header']

        # Called for each request that goes through the downloader
        # middleware.

        # Must either:
        # - return None: continue processing this request
        # - or return a Response object
        # - or return a Request object
        # - or raise IgnoreRequest: process_exception() methods of
        #   installed downloader middleware will be called
        return None

    def process_response(self, request, response, spider):
        # Called with the response returned from the downloader.

        # Must either;
        # - return a Response object
        # - return a Request object
        # - or raise IgnoreRequest

        headers = request.headers
        proxy_try_count = self.get_proxy_try_count_from_request(request)
        if response.status != 200 and proxy_try_count < 10:
            headers["Proxy_Try_Count"] = str(proxy_try_count)
            headers["Proxy_URL"] = self.get_proxy_from_request(request)
            request.replace(headers=headers)
            return request

        return response

    def process_exception(self, request, exception, spider):
        # Called when a download handler or a process_request()
        # (from other downloader middleware) raises an exception.

        # Must either:
        # - return None: continue processing this exception
        # - return a Response object: stops process_exception() chain
        # - return a Request object: stops process_exception() chain
        pass

    def spider_opened(self, spider):
        spider.logger.info('Spider opened: %s' % spider.name)

    def get_proxy(self, request, name, url):
        proxy_try_count = request.headers.get('Proxy_Try_Count')
        proxy_url = request.headers.get('Proxy_URL')
        try:
            response = requests.get(self.proxy_server_addr + f"/get?name={name}&url={url}", timeout=10,
                                    headers={"proxy_try_count": proxy_try_count, "proxy_url": proxy_url})
        except Exception as err:
            print(f'get proxy error occurred: {err}, spider name: {name}, request url: {url}')
            return None

        resp = response.json()

        return {
            'host': resp['host'],
            'user': resp['user'],
            'pass': resp['pass'],
            'port': resp['port'],
            'type': resp['type'],
            'basic_auth_header': resp['basic_auth_header'],
        }

    def get_proxy_from_request(self, request):
        if 'splash' in request.meta:
            return request.meta['splash']['args']['proxy'];
        else:
            return request.meta['proxy']

    def get_proxy_try_count_from_request(self, request):
        headers = request.headers
        if headers.get('Proxy_Try_Count') is None:
            proxy_try_count = 1
        else:
            proxy_try_count = int(headers.get('Proxy_Try_Count'))
            proxy_try_count += 1

        return proxy_try_count
